(*--------------------------------------------------------------------------+
 |                                                                          |
 |                   Bote  outils pour les arbres gnraux                |
 |                                                                          |
 |  M. Quercia, le 06/08/96                                                 |
 +--------------------------------------------------------------------------+*)


(* applique une fonction  un arbre et renvoie l'arbre des rsultats *)
let rec b_map f = function
| B_vide -> B_vide
| B_noeud(x,g,d) -> B_noeud(f(x), b_map f g, b_map f d);;


        (* +---------------------------------------------------------+
           |  Lecture d'une expression parenthse et mise en arbre  |
           +---------------------------------------------------------+ *)


type terme = Pg | Pd | Eos | Terme of string;;

let rec mot_suivant(ch,i) =
  let l = string_length(ch) in
  if i >= l then (Eos,i)
  else match ch.[i] with
  | (` `|`\t`|`\n`) -> mot_suivant(ch,i+1)
  | `(` -> (Pg,i+1)
  | `)` -> (Pd,i+1)
  | _ -> let j = ref(i+1) and k = ref(i+1) and encore = ref(true) in
         while !j < l & !encore do match ch.[!j] with
         | (` `|`\t`|`\n`) -> incr j
	 | (`(`| `)`)      -> encore := false
         | _               -> incr j; k := !j
         done;
         (Terme(sub_string ch i (!k - i)), !j)
;; 


let erreur(msg,ch,i) =
  prerr_endline("Erreur : " ^ msg);
  prerr_endline(ch);
  prerr_endline((make_string i ` `) ^ "^");
  failwith "arbre mal form"
;;

let b_arbre =

  let rec arbre(ch,i) = match mot_suivant(ch,i) with
  | (Pg,j) ->
    begin
      match mot_suivant(ch,j) with
      | (Terme(e),k) ->
        begin
          let (g,l) = arbre(ch,k) in
          let (d,m) = arbre(ch,l) in
          match mot_suivant(ch,m) with
          | (Pd,n) -> (B_noeud(e,g,d),n)
          | _ -> erreur("il manque \")\"", ch,m)
        end
      | (Pd,k) -> (B_vide,k)
      | _ -> erreur("il manque un terme", ch,j)
    end
  | _ -> (B_vide,i)
  in

  fun ch -> let (a,i) = arbre(ch,0) in
            match mot_suivant(ch,i) with
            | (Eos,_) -> a
            | _ -> erreur("expression hors parenthses",ch,i)
;;


                (* +-----------------------------------------+
                   |  Imprime un arbre sous forme graphique  |
                   +-----------------------------------------+ *)

(* calcule les abscisses relatives d'impression  *)
(* des noeuds et l'encombrement total d'un arbre *)
let rec abscisses_arbre = function

| B_vide -> (0,B_vide)

| B_noeud(tiq,B_vide,B_vide) ->
   let l = string_length(tiq) in (l, B_noeud((l/2,tiq),B_vide,B_vide))

| B_noeud(tiq,g,B_vide) ->
   let l = string_length(tiq)
   and (eg,xg,g') = match abscisses_arbre g with
   | (eg,B_noeud((x,e),g,d)) -> (eg,x,B_noeud((-1,e),g,d))
   | _ -> failwith "cas impossible"
   in
   let x0 = min (-l/2) (-xg-1) and x1 = max (l-l/2) (eg-xg-1)
   in (x1-x0, B_noeud((-x0,tiq),g',B_vide))

| B_noeud(tiq,B_vide,d) ->
   let l = string_length(tiq)
   and (ed,xd,d') = match abscisses_arbre d with
   | (ed,B_noeud((x,e),g,d)) -> (ed,x,B_noeud((1,e),g,d))
   | _ -> failwith "cas impossible"
   in
   let x0 = min (-l/2) (-xd+1) and x1 = max (l-l/2) (ed-xd+1)
   in (x1-x0, B_noeud((-x0,tiq),B_vide,d'))

| B_noeud(tiq,g,d) ->

   let (eg,xg,tig,gg,gd) = match abscisses_arbre g with
   | (eg,B_noeud((x,tig),gg,gd)) -> (eg,x,tig,gg,gd)
   | _ -> failwith "cas impossible"
   in
  
   let (ed,xd,tid,dg,dd) = match abscisses_arbre d with
   | (ed,B_noeud((x,tid),dg,dd)) -> (ed,eg+x+2,tid,dg,dd)
   | _ -> failwith "cas impossible"
   in

   let xm = (xd+xg)/2 in
   let tiq' = centre tiq (xd-xg-3) in
   let l = string_length(tiq')
   in (max (eg+ed+2) l, B_noeud((max xm (l/2),tiq'),
                          B_noeud((xg-xm,tig),gg,gd),
                          B_noeud((xd-xm,tid),dg,dd)))

and centre chaine l =
  let lc = string_length(chaine) in
  if lc >= l then chaine
             else let g = (l-lc)/2 in let d = l-lc-g in
                  (make_string g `_`) ^ chaine ^(make_string d `_`)

;;


(* impression *)
#open "format";;

let rec print_noeuds x0 = function
| [] -> if x0 > 0 then print_cut()
| B_vide::suite -> print_noeuds x0 suite
| B_noeud((x,e),_,_)::suite ->
    let l = string_length(e) in
    for i=x0 to x-(l-1)/2-1 do print_char ` ` done;
    print_string(e);
    print_noeuds (x+l/2+1) suite
;;

let rec print_traits_fils x0 x = function
| B_vide -> x0
| B_noeud((y,_),_,_) ->
    let z = (if y < 0 then x+y else if y = 0 then x+y-1 else x+y-2) in
    for i=x0 to z do print_char ` ` done;
    print_char (if y < 0 then `/` else if y = 0 then `|` else `\\`);
    z+2
;;

let rec print_traits x0 = function
| [] -> if x0 > 0 then print_cut()
| B_vide::suite -> print_traits x0 suite
| B_noeud((x,_),g,d)::suite ->
    let x1 = print_traits_fils x0 x g in
    let x2 = print_traits_fils x1 x d in
    print_traits x2 suite
;;

let rec liste_fils = function
| [] -> []
| B_vide::suite -> liste_fils suite
| B_noeud((x,e),g,d)::suite ->
   (match g with B_vide -> [] | B_noeud((y,e),g,d) -> [B_noeud((x+y,e),g,d)])
 @ (match d with B_vide -> [] | B_noeud((y,e),g,d) -> [B_noeud((x+y,e),g,d)])
 @ (liste_fils suite)
;;

let rec print_fort = function
| [] -> ()
| f  -> print_noeuds 0 f;
        print_traits 0 f;
        print_fort(liste_fils f)
;;

let print_b_arbre a =
  open_vbox(0);
  print_fort [snd(abscisses_arbre a)];
  close_box()
;;

#close "format";;



                               (* +----------+
                                  |  Divers  |
                                  +----------+ *)

let print_int_b_arbre(a) = print_b_arbre (b_map string_of_int a)
and int_b_arbre(ch) = b_map int_of_string (b_arbre ch)
;;
