(*--------------------------------------------------------------------------+
 |                                                                          |
 |                   Bote  outils pour les arbres gnraux                |
 |                                                                          |
 |  M. Quercia, le 06/08/96                                                 |
 +--------------------------------------------------------------------------+*)


(* applique une fonction  un arbre et renvoie l'arbre des rsultats *)
let rec g_map f = function
  G_noeud(x,fils) -> G_noeud(f(x), map (g_map f) fils);;


        (* +---------------------------------------------------------+
           |  Lecture d'une expression parenthse et mise en arbre  |
           +---------------------------------------------------------+ *)


type terme = Pg | Pd | Eos | Terme of string;;

let rec mot_suivant(ch,i) =
  let l = string_length(ch) in
  if i >= l then (Eos,i)
  else match ch.[i] with
  | (` `|`\t`|`\n`) -> mot_suivant(ch,i+1)
  | `(` -> (Pg,i+1)
  | `)` -> (Pd,i+1)
  | _ -> let j = ref(i+1) and k = ref(i+1) and encore = ref(true) in
         while !j < l & !encore do match ch.[!j] with
         | (` `|`\t`|`\n`) -> incr j
	 | (`(`| `)`)      -> encore := false
         | _               -> incr j; k := !j
         done;
         (Terme(sub_string ch i (!k - i)), !j)
;; 


let erreur(msg,ch,i) =
  prerr_endline("Erreur : " ^ msg);
  prerr_endline(ch);
  prerr_endline((make_string i ` `) ^ "^");
  failwith "arbre mal form"
;;

let g_arbre =

  let rec arbre(ch,i) = match mot_suivant(ch,i) with
  | (Pg,j) -> begin
                match mot_suivant(ch,j) with
                | (Terme(e),k) -> let (f,l) = fort(ch,k) in (G_noeud(e,f),l)
                | _ -> erreur("il manque un terme", ch,j)
              end
  | _ -> erreur("il manque \"(\"",ch,i)

  and fort(ch,i) = match mot_suivant(ch,i) with
  | (Pd,j) -> ([],j)
  | _ -> let (a,j) = arbre(ch,i) in
         let (f,k) = fort(ch,j) in (a::f,k)
  in

  fun ch -> let (a,i) = arbre(ch,0) in
            match mot_suivant(ch,i) with
            | (Eos,_) -> a
            | _ -> erreur("expression hors parenthses",ch,i)
;;


                (* +-----------------------------------------+
                   |  Imprime un arbre sous forme graphique  |
                   +-----------------------------------------+ *)

(* calcule les abscisses relatives d'impression  *)
(* des noeuds et l'encombrement total d'un arbre *)
let rec abscisses_arbre x0 = function G_noeud(tiq,fils) ->

  let (enc,xf0,xfn,fort) = abscisses_fort 0 fils in
  let xf = (xf0+xfn)/2 in
  let fort' = dcale (-xf) fort in

  let tiq' = centre tiq (xfn-xf0-3) in
  let l = string_length(tiq') in
  (max enc l, G_noeud((x0+(max xf (l/2)),tiq'),fort'))

and abscisses_fort x0 = function
| []  -> (0,0,0,[])
| [a] -> let (enc,(G_noeud((x,_),_) as b)) = abscisses_arbre x0 a in (enc,x,x,[b])
| a::suite ->
    let (enc,(G_noeud((x,_),_) as b)) = abscisses_arbre x0 a in
    let (enc',_,y,f) = abscisses_fort (x0+enc+2) suite in (enc+enc'+2,x,y,b::f)

and centre chaine l =
  let lc = string_length(chaine) in
  if lc >= l then chaine
             else let g = (l-lc)/2 in let d = l-lc-g in
                  (make_string g `_`) ^ chaine ^(make_string d `_`)


and dcale dx = function
| [] -> []
| G_noeud((x,e),f)::suite -> G_noeud((x+dx,e),f) :: dcale dx suite

;;


(* impression *)
#open "format";;
let rec print_noeuds x0 = function
| [] -> if x0 > 0 then print_cut()
| G_noeud((x,e),_)::suite ->
    let l = string_length(e) in
    for i=x0 to x-(l-1)/2-1 do print_char ` ` done;
    print_string(e);
    print_noeuds (x+l/2+1) suite
;;

let rec print_traits_fils x0 x = function
| [] -> x0
| G_noeud((y,_),_)::suite ->
    let z = (if y < 0 then x+y else if y = 0 then x+y-1 else x+y-2) in
    for i=x0 to z do print_char ` ` done;
    print_char (if y < 0 then `/` else if y = 0 then `|` else `\\`);
    print_traits_fils (z+2) x suite
;;

let rec print_traits x0 = function
| [] -> if x0 > 0 then print_cut()
| G_noeud((x,_),f)::suite ->
    let x1 = print_traits_fils x0 x f in
    print_traits x1 suite
;;

let rec liste_fils = function
| [] -> []
| G_noeud((x,e),f)::suite -> (dcale x f) @ (liste_fils suite)
;;

let rec print_fort = function
| [] -> ()
| f  -> print_noeuds 0 f;
        print_traits 0 f;
        print_fort(liste_fils f)
;;

let print_g_arbre a =
  open_vbox(0);
  print_fort [snd(abscisses_arbre 0 a)];
  close_box()
;;

#close "format";;


                               (* +----------+
                                  |  Divers  |
                                  +----------+ *)

let print_int_g_arbre(a) = print_g_arbre (g_map string_of_int a)
and int_g_arbre(ch) = g_map int_of_string (g_arbre ch)
;;
